<?php
/**
 * Overrides carrier shipping with Table Rate Shipping
 *
 * Table Rate Shipping by Kahanit(https://www.kahanit.com/) is licensed under a
 * Creative Creative Commons Attribution-NoDerivatives 4.0 International License.
 * Based on a work at https://www.kahanit.com/.
 * Permissions beyond the scope of this license may be available at https://www.kahanit.com/.
 * To view a copy of this license, visit http://creativecommons.org/licenses/by-nd/4.0/.
 *
 * @author    Amit Sidhpura <amit@kahanit.com>
 * @copyright 2015 Kahanit
 * @license   http://creativecommons.org/licenses/by-nd/4.0/
 */

require_once dirname(__FILE__) . '/../../libraries/TRSCarrierTableRate.php';
require_once dirname(__FILE__) . '/../../libraries/TRSCSV.php';
require_once dirname(__FILE__) . '/../../libraries/TRSUploadHandler.php';

/**
 * Class AdminTableRateShippingController
 */
class AdminTableRateShippingController extends ModuleAdminController
{
    public function init()
    {
        parent::init();

        if (!Shop::getContextShopID()) {
            $controller = 'AdminTableRateShipping';
            $id_lang = $this->context->language->id;
            $params = array(
                'token' => Tools::getAdminTokenLite($controller),
                'setShopContext' => 's-' . current(Shop::getContextListShopID())
            );
            $link = Dispatcher::getInstance()->createUrl($controller, $id_lang, $params, false);
            Tools::redirectAdmin($link);
            die();
        }

        $this->bootstrap = true;
    }

    public function setMedia()
    {
        parent::setMedia();

        $this->addCSS('https://cdn.datatables.net/t/bs/dt-1.10.11/datatables.min.css');
        $this->addCSS('https://cdnjs.cloudflare.com/ajax/libs/select2/4.0.2/css/select2.min.css');
        $this->addCSS('https://cdnjs.cloudflare.com/ajax/libs/blueimp-file-upload/9.12.1/css/jquery.fileupload.min.css');
        $this->addCSS('https://cdnjs.cloudflare.com/ajax/libs/blueimp-file-upload/9.12.1/css/jquery.fileupload-ui.min.css');
        $this->addCSS(_MODULE_DIR_ . $this->module->name . '/views/css/bootstrap-editable.css');
        $this->addCSS(_MODULE_DIR_ . $this->module->name . '/views/css/admin.css');

        $this->addJquery();
        $this->addJqueryUI('ui.widget');
        $this->addJqueryUI('ui.tabs');
        $this->addJqueryUI('ui.sortable');
        $this->addJS('https://cdn.datatables.net/t/bs/dt-1.10.11/datatables.min.js');
        $this->addJS('https://cdnjs.cloudflare.com/ajax/libs/select2/4.0.2/js/select2.full.min.js');
        $this->addJS('https://cdnjs.cloudflare.com/ajax/libs/blueimp-JavaScript-Templates/3.4.0/js/tmpl.min.js');
        $this->addJS('https://cdnjs.cloudflare.com/ajax/libs/blueimp-file-upload/9.12.1/js/jquery.iframe-transport.min.js');
        $this->addJS('https://cdnjs.cloudflare.com/ajax/libs/blueimp-file-upload/9.12.1/js/jquery.fileupload.min.js');
        $this->addJS('https://cdnjs.cloudflare.com/ajax/libs/blueimp-file-upload/9.12.1/js/jquery.fileupload-process.min.js');
        $this->addJS('https://cdnjs.cloudflare.com/ajax/libs/blueimp-file-upload/9.12.1/js/jquery.fileupload-validate.min.js');
        $this->addJS('https://cdnjs.cloudflare.com/ajax/libs/blueimp-file-upload/9.12.1/js/jquery.fileupload-ui.min.js');
        $this->addJS(_MODULE_DIR_ . $this->module->name . '/views/js/bootstrap-editable.min.js');
        $this->addJS('https://cdnjs.cloudflare.com/ajax/libs/jquery.serializeJSON/2.7.2/jquery.serializejson.min.js');
        $this->addJS(_MODULE_DIR_ . $this->module->name . '/views/js/jquery.admintablerateshipping.js');
    }

    public function initContent()
    {
        parent::initContent();

        $ajax = (bool)Tools::getValue('ajax', false);
        if ($ajax) {
            return;
        }

        $this->context->smarty->assign(array(
            'default_shop' => $this->context->shop->id,
            'module_views_dir' => _PS_ROOT_DIR_ . '/modules/' . $this->module->name . '/views/',
            'module_name' => $this->module->name,
            'module_display_name' => $this->module->displayName,
            'module_version' => $this->module->version,
            'carriers' => Carrier::getCarriers($this->context->language->id, true, false, false, null, ALL_CARRIERS),
            'zones' => array_merge([['id_zone' => 0, 'name' => 'All Zones']], Zone::getZones()),
            'countries' => Country::getCountries($this->context->language->id),
            'csv_separator' => Configuration::get('TABLERATESHIPPING_CSV_SEPARATOR')
        ));

        $this->setTemplate('configuration.tpl');
    }

    public function displayAjax()
    {
        $method = Tools::getValue('method', false);
        $header = Tools::getValue('header', 'json');

        if ($method !== false) {
            if ($header == 'json') {
                header('Content-Type: application/json');
                echo Tools::jsonEncode($this->$method());
            } else {
                echo $this->$method();
            }
        }

        die();
    }

    public function getRuleGroups()
    {
        $start = (int)Tools::getValue('start', 0);
        $length = (int)Tools::getValue('length', 20);
        $id_carrier = (int)Tools::getValue('id_carrier', 0);
        $id_carrier = self::getCarrierIdFromCarrierId($id_carrier);
        $id_zone = (int)Tools::getValue('id_zone', -1);
        $id_lang = $this->context->language->id;
        $id_shop = $this->context->shop->id;
        $where = '';

        $records = TRSCarrierTableRate::getRuleGroups($id_carrier, $id_zone, $start, $length, $where, $id_lang,
            $id_shop);
        $total_records = TRSCarrierTableRate::getRuleGroupsNumRows($id_carrier, $id_zone, $where, $id_shop);

        return array(
            'recordsTotal' => $total_records,
            'recordsFiltered' => $total_records,
            'data' => $records
        );
    }

    public function getRules()
    {
        $start = (int)Tools::getValue('start', 0);
        $length = (int)Tools::getValue('length', 20);

        $id_carrier = (int)Tools::getValue('id_carrier', 0);
        $id_carrier = self::getCarrierIdFromCarrierId($id_carrier);
        $id_zone = (int)Tools::getValue('id_zone', -1);
        $id_country = (int)Tools::getValue('id_country', 0);
        $id_state = (int)Tools::getValue('id_state', 0);
        $dest_city = Tools::getValue('dest_city', '');
        $id_shop = $this->context->shop->id;
        $where = '';

        $records = TRSCarrierTableRate::getRules($id_carrier, $id_zone, $id_country, $id_state, $dest_city, $start,
            $length, $where, $id_shop);
        $total_records = TRSCarrierTableRate::getRulesNumRows($id_carrier, $id_zone, $id_country, $id_state, $dest_city,
            $where, $id_shop);

        return array(
            'recordsTotal' => $total_records,
            'recordsFiltered' => $total_records,
            'data' => $records
        );
    }

    public function getProducts()
    {
        $order = Tools::getValue('order', array(array('column' => 0, 'dir' => 'ASC')));
        $columns = Tools::getValue('columns', array());

        $start = (int)Tools::getValue('start', 0);
        $length = (int)Tools::getValue('length', 25);
        $id_carrier = (int)Tools::getValue('id_carrier', 0);
        $id_carrier = self::getCarrierIdFromCarrierId($id_carrier);
        $category_id = (int)Tools::getValue('id_category_default', null);
        $search_get = Tools::getValue('search', array('value' => ''));
        $search = $search_get['value'];

        $orderfld = Tools::strtolower($columns[$order[0]['column']]['data']);
        $orderfld = (($orderfld == 'name') ? 'l.' . $orderfld : 'p.' . $orderfld);
        $orderdir = Tools::strtoupper($order[0]['dir']);

        $lang_id = (int)Tools::getValue('id_lang', $this->context->language->id);

        $products = $this->getProductsModel($lang_id, $category_id, $search, $start, $length,
            $orderfld, $orderdir, $this->context->shop->id);
        $products_total = $this->getProductsModelNumRows($category_id, $search, $this->context->shop->id);
        $records = array();

        foreach ($products as $product) {
            $producttemp = new Product($product['id_product'], false, $lang_id);
            $categorytempid = ((!$producttemp->id_category_default)
                ? $this->context->shop->id_category
                : $producttemp->id_category_default);
            $categorytemp = new Category($categorytempid, $lang_id);

            $records[] = array(
                'id_product' => $producttemp->id,
                'name' => $producttemp->name,
                'reference' => $producttemp->reference,
                'id_category_default' => $categorytemp->id,
                'name_category' => $categorytemp->name,
                'active' => $producttemp->active == '1',
                'hascarrier' => $this->hasProductCarrier($producttemp->id, $id_carrier)
            );
        }

        return array(
            'recordsTotal' => $products_total,
            'recordsFiltered' => $products_total,
            'data' => $records
        );
    }

    public function getZonesOrder()
    {
        $id_shop = $this->context->shop->id;
        $id_carrier = (int)Tools::getValue('id_carrier', 0);
        $id_carrier = self::getCarrierIdFromCarrierId($id_carrier);
        $order = [];

        if ($id_carrier !== 0) {
            $sql = new DbQuery();
            $sql->select('DISTINCT id_zone');
            $sql->from('carrier_table_rate');
            $sql->where('id_carrier = ' . (int)$id_carrier . ' AND id_shop = ' . (int)$id_shop);
            $sql->orderBy('`order` ASC');
            $results = Db::getInstance()->executeS($sql);
            foreach ($results as $result) {
                $order[] = $result['id_zone'];
            }
        }

        return $order;
    }

    public function save()
    {
        $id_lang = $this->context->language->id;
        $id_shop = $this->context->shop->id;
        $action = Tools::getValue('action', '');
        $id_carrier = (int)Tools::getValue('id_carrier', 0);
        $id_carrier = self::getCarrierIdFromCarrierId($id_carrier);
        $records = Tools::getValue('records', []);

        if ($id_carrier === 0) {
            return [
                'status' => 'danger',
                'message' => 'Invalid carrier or zone'
            ];
        }

        return TRSCarrierTableRate::saveRules($action, $id_carrier, $records, $id_lang, $id_shop);
    }

    public function updateCountryStateCity()
    {
        $update = Tools::getValue('update', false);
        $id_lang = $this->context->language->id;
        $id_shop = $this->context->shop->id;
        $id_carrier = (int)Tools::getValue('id_carrier', 0);
        $id_carrier = self::getCarrierIdFromCarrierId($id_carrier);
        $id_zone = (int)Tools::getValue('id_zone', -1);
        $id_country = (int)Tools::getValue('id_country', 0);
        $id_state = (int)Tools::getValue('id_state', 0);
        $dest_city = Tools::getValue('dest_city', false);
        $value = Tools::getValue('value', false);

        // set new data variables
        if ($update == 'id_country') {
            $id_country_new = $value;
        } else {
            $id_country_new = $id_country;
        }
        if ($update == 'id_state') {
            $id_state_new = $value;
        } else {
            $id_state_new = $id_state;
        }
        if ($update == 'dest_city') {
            $dest_city_new = $value;
        } else {
            $dest_city_new = $dest_city;
        }

        // reset state of country changes
        if ($update == 'id_country') {
            $id_state_new = 0;
        }

        if ($update !== false && in_array($update,
                ['id_country', 'id_state', 'dest_city']) && $id_carrier !== 0 && $id_zone !== -1
            && $id_country_new !== false && $id_state_new !== false && $dest_city_new !== false
        ) {
            // validate group already exists
            $rules = TRSCarrierTableRate::getRules($id_carrier, $id_zone, $id_country, $id_state, $dest_city, null,
                null, '', $id_shop);
            foreach ($rules as $rule) {
                $id_carrier_table_rate_explode = array_map('intval', explode(',', $rule['id_carrier_table_rate']));
                foreach ($id_carrier_table_rate_explode as $id_carrier_table_rate_explode_item) {
                    $carrier_table_rate = new TRSCarrierTableRate($id_carrier_table_rate_explode_item, $id_lang);
                    $carrier_table_rate->id_country = (int)(($id_country_new == '*') ? 0 : $id_country_new);
                    $carrier_table_rate->id_state = (int)(($id_state_new == '*') ? 0 : $id_state_new);
                    $carrier_table_rate->dest_city = pSQL(($dest_city_new == '*') ? '' : $dest_city_new);
                    $message = $carrier_table_rate->validateFields(false, true);
                    if ($message != 1) {
                        return [
                            'status' => 'danger',
                            'message' => $message
                        ];
                    }
                }
            }

            // update values
            Db::getInstance()->update('carrier_table_rate', [
                'id_country' => (int)(($id_country_new == '*') ? 0 : $id_country_new),
                'id_state' => (int)(($id_state_new == '*') ? 0 : $id_state_new),
                'dest_city' => pSQL(($dest_city_new == '*') ? '' : $dest_city_new)
            ], 'id_carrier = ' . (int)$id_carrier . '
                AND id_zone = ' . (int)$id_zone . '
                AND id_country = ' . (int)(($id_country == '*') ? 0 : $id_country) . '
                AND id_state = ' . (int)(($id_state == '*') ? 0 : $id_state) . '
                AND dest_city = \'' . pSQL((($dest_city == '*') ? '' : $dest_city)) . '\'
                AND id_shop = ' . (int)$id_shop);

            if ((int)Db::getInstance()->getValue('SELECT ROW_COUNT()') > 0) {
                $country = new Country($id_country_new, $this->context->language->id, $id_shop);
                $state = new State($id_state_new, $this->context->language->id, $id_shop);
                return [
                    'status' => 'success',
                    'id_shop' => $id_shop,
                    'id_carrier' => $id_carrier,
                    'id_zone' => $id_zone,
                    'id_country' => ($id_country_new == 0) ? '*' : $id_country_new,
                    'country' => ($country->name && $country->iso_code) ? $country->name . ' (' . $country->iso_code . ')' : '*',
                    'id_state' => ($id_state_new == 0) ? '*' : $id_state_new,
                    'state' => ($state->name && $state->iso_code) ? $state->name . ' (' . $state->iso_code . ')' : '*',
                    'dest_city' => (($dest_city_new == '') ? '*' : $dest_city_new),
                    'message' => 'Record updated successfully'
                ];
            }
        }

        return [
            'status' => 'danger',
            'message' => 'Error updating record'
        ];
    }

    public function updateConditionPriceComment()
    {
        $update = Tools::getValue('update', false);
        $id_lang = $this->context->language->id;
        $id_carrier_table_rate = Tools::getValue('id_carrier_table_rate', false);
        $value = Tools::getValue('value', false);
        $fields = [
            'condition_weight',
            'condition_price',
            'condition_ptprice',
            'condition_quantity',
            'condition_volume',
            'price',
            'comment'
        ];
        $condition_types = ['weight', 'price', 'ptprice', 'quantity', 'volume'];

        if ($update !== false && in_array($update, $fields) && $id_carrier_table_rate !== false) {
            if ($update == 'price') {
                $update_values = [
                    $update => pSQL($value, true)
                ];
            } elseif ($update == 'comment') {
                $update_values = [
                    $update => str_replace('\"', '"', pSQL($value))
                ];
            } else {
                if (is_numeric($value)) {
                    $value_from = $value_to = $value;
                } else {
                    $value_range = explode('-', $value);
                    if (isset($value_range[0]) && is_numeric($value_range[0]) && isset($value_range[1]) && is_numeric($value_range[1])) {
                        $value_from = $value_range[0];
                        $value_to = $value_range[1];
                    } else {
                        $value_from = $value_to = -1;
                    }
                }
                $update_values = [
                    $update . '_from' => (float)(($value_from == '*') ? -1 : $value_from),
                    $update . '_to' => (float)(($value_to == '*') ? -1 : $value_to)
                ];
            }

            // validate records
            $id_carrier_table_rate_explode = array_map('intval', explode(',', $id_carrier_table_rate));
            foreach ($id_carrier_table_rate_explode as $id_carrier_table_rate_explode_item) {
                $carrier_table_rate = new TRSCarrierTableRate($id_carrier_table_rate_explode_item, $id_lang);
                foreach ($update_values as $update_value_key => $update_value_value) {
                    $carrier_table_rate->{$update_value_key} = $update_value_value;
                }
                $message = $carrier_table_rate->validateFields(false, true);
                if ($message != 1) {
                    return [
                        'status' => 'danger',
                        'message' => $message
                    ];
                }
            }

            $result = Db::getInstance()->update('carrier_table_rate', $update_values,
                'id_carrier_table_rate IN (' . $id_carrier_table_rate . ')');

            if ($result) {
                $id_carrier_table_rate = array_map('intval', explode(',', $id_carrier_table_rate));
                if (isset($id_carrier_table_rate[0])) {
                    $carrier_table_rate = new TRSCarrierTableRate((int)$id_carrier_table_rate[0]);
                    $status = [
                        'status' => 'success',
                        'price' => $carrier_table_rate->price,
                        'comment' => $carrier_table_rate->comment,
                        'message' => 'Record updated successfully'
                    ];
                    foreach ($condition_types as $condition_type) {
                        $status['condition_' . $condition_type . '_from'] = ($carrier_table_rate->{'condition_' . $condition_type . '_from'} == -1)
                            ? '*' : $carrier_table_rate->{'condition_' . $condition_type . '_from'};
                        $status['condition_' . $condition_type . '_to'] = ($carrier_table_rate->{'condition_' . $condition_type . '_to'} == -1)
                            ? '*' : $carrier_table_rate->{'condition_' . $condition_type . '_to'};
                    }
                    return $status;
                }
            }
        }

        return [
            'status' => 'danger',
            'message' => 'Error updating record'
        ];
    }

    public function updateStatus()
    {
        $id_carrier_table_rate = Tools::getValue('id_carrier_table_rate', false);
        $active = (int)Tools::getValue('active', 1);

        if ($id_carrier_table_rate !== false) {
            Db::getInstance()->update('carrier_table_rate', [
                'active' => (int)$active
            ], 'id_carrier_table_rate IN (' . $id_carrier_table_rate . ')');

            return array(
                'status' => 'success',
                'message' => 'Status updated successfully',
                'id_carrier_table_rate' => $id_carrier_table_rate,
                'active' => $active
            );
        }

        return array(
            'status' => 'danger',
            'message' => 'Status update failed',
            'id_carrier_table_rate' => $id_carrier_table_rate,
            'active' => $active
        );
    }

    public function updateOrder()
    {
        $id_shop = $this->context->shop->id;
        $entity = Tools::getValue('entity', '');
        $id_carrier = (int)Tools::getValue('id_carrier', 0);
        $id_carrier = self::getCarrierIdFromCarrierId($id_carrier);
        $id_zone = (int)Tools::getValue('id_zone', -1);
        $order = Tools::getValue('order', []);

        return TRSCarrierTableRate::updateOrder($entity, $id_carrier, $id_zone, $order, $id_shop);
    }

    public function updateProductSelected()
    {
        $id_product = (int)Tools::getValue('id_product', 0);
        $id_carrier = (int)Tools::getValue('id_carrier', 0);
        $id_carrier = self::getCarrierIdFromCarrierId($id_carrier);
        $select = (int)Tools::getValue('select', 1);

        if ($id_product !== 0 && $id_carrier !== 0) {
            if ($select && !$this->hasProductCarrier($id_product, $id_carrier)) {
                $carrier = new Carrier($id_carrier);

                Db::getInstance()->insert('product_carrier', array(
                    'id_product' => (int)$id_product,
                    'id_carrier_reference' => (int)$carrier->id_reference,
                    'id_shop' => (int)$this->context->shop->id
                ));

                return array(
                    'status' => 'success',
                    'message' => 'Product selected',
                    'id_product' => $id_product,
                    'id_carrier' => $id_carrier,
                    'select' => $this->hasProductCarrier($id_product, $id_carrier)
                );
            }

            if (!$select && $this->hasProductCarrier($id_product, $id_carrier)) {
                $carrier = new Carrier($id_carrier);

                Db::getInstance()->delete('product_carrier',
                    'id_product = ' . (int)$id_product . '
                        AND id_carrier_reference = ' . (int)$carrier->id_reference . '
                        AND id_shop = ' . (int)$this->context->shop->id);

                return array(
                    'status' => 'success',
                    'message' => 'Product deselected',
                    'id_product' => $id_product,
                    'id_carrier' => $id_carrier,
                    'select' => $this->hasProductCarrier($id_product, $id_carrier)
                );
            }
        }

        return array(
            'status' => 'danger',
            'message' => 'Product select failed',
            'id_product' => $id_product,
            'id_carrier' => $id_carrier,
            'select' => $select
        );
    }

    public function delete()
    {
        $type = Tools::getValue('type', '');
        $entity = Tools::getValue('entity', '');
        $id_carrier_table_rate = Tools::getValue('id_carrier_table_rate', false);
        $id_shop = $this->context->shop->id;
        $id_carrier = (int)Tools::getValue('id_carrier', 0);
        $id_carrier = self::getCarrierIdFromCarrierId($id_carrier);
        $id_zone = (int)Tools::getValue('id_zone', -1);
        $data = Tools::getValue('data', false);

        if ($entity == 'rule') {
            if ($id_carrier_table_rate !== false && ($type == 'single' || $type == 'selected')) {
                $id_carrier_table_rate = array_map('intval', explode(',', $id_carrier_table_rate));
                if (isset($id_carrier_table_rate[0]) && is_numeric($id_carrier_table_rate[0])) {
                    $carrier_table_rate = new TRSCarrierTableRate((int)$id_carrier_table_rate[0]);
                    $id_carrier = $carrier_table_rate->id_carrier;
                }
                $id_carrier_table_rate = implode(',', $id_carrier_table_rate);
                Db::getInstance()->delete('carrier_table_rate',
                    'id_carrier_table_rate IN (' . $id_carrier_table_rate . ')');
            }
        } elseif ($entity == 'rulegroup') {
            if ($data !== false && ($type == 'single' || $type == 'selected')) {
                foreach ($data as $line) {
                    Db::getInstance()->delete('carrier_table_rate',
                        'id_carrier = ' . (int)$id_carrier . ' AND id_zone = ' . (int)$line['id_zone'] . '
                    AND id_country = ' . (int)(($line['id_country'] == '*') ? 0 : $line['id_country']) . '
                    AND id_state = ' . (int)(($line['id_state'] == '*') ? 0 : $line['id_state']) . '
                    AND dest_city = \'' . pSQL((($line['dest_city'] == '*') ? '' : $line['dest_city'])) . '\'
                    AND id_shop = ' . (int)$id_shop);
                }
            } elseif ($id_zone !== -1 && $type == 'all') {
                Db::getInstance()->delete('carrier_table_rate',
                    'id_zone = ' . (int)$id_zone . ' AND id_shop = ' . (int)$id_shop);
            }
        }

        // clear carrier as module carrier if no records
        $records_total = TRSCarrierTableRate::getRuleGroupsNumRows($id_carrier, null, '', false);
        if ($records_total == 0) {
            TRSCarrierTableRate::clearCarrierAsTableRate($id_carrier);
        }

        return array(
            'status' => 'success',
            'message' => 'Table rates deleted successfully'
        );
    }

    public function processAttachment()
    {
        $id_employee = $this->context->cookie->id_employee;
        $token = Tools::getAdminToken('AdminTableRateShipping' . (int)Tab::getIdFromClassName('AdminTableRateShipping') . (int)$id_employee);
        $url = 'index.php?controller=AdminTableRateShipping&token=' . $token . '&ajax=1';

        $uploads = new TRSUploadHandler(array(
            'accept_file_types' => '/(\.|\/)(csv)$/i',
            'upload_dir' => _PS_ADMIN_DIR_ . '/import/' . $this->module->name . '/',
            'upload_url' => $url . '&method=downloadAttachment&header=csv&file=',
            'script_url' => $url . '&method=processAttachment',
            'print_response' => false
        ));

        $response = $uploads->get_response();

        return $response;
    }

    public function downloadAttachment()
    {
        $file = Tools::getValue('file', '');

        if ($file == '') {
            return;
        }

        $file = _PS_ADMIN_DIR_ . '/import/' . $this->module->name . '/' . $file;

        header('Content-Description: File Transfer');
        header('Content-Type: application/octet-stream');
        header('Content-Disposition: attachment; filename=' . basename($file));
        header('Expires: 0');
        header('Cache-Control: must-revalidate');
        header('Pragma: public');
        header('Content-Length: ' . filesize($file));

        readfile($file);
        exit;
    }

    public function import()
    {
        $file = Tools::getValue('file', '');
        $csv_separator = Tools::getValue('csv_separator', ',');
        $id_lang = $this->context->language->id;
        $id_shop = $this->context->shop->id;
        $id_carrier = (int)Tools::getValue('id_carrier', 0);
        $id_carrier = self::getCarrierIdFromCarrierId($id_carrier);
        $file = _PS_ADMIN_DIR_ . '/import/' . $this->module->name . '/' . $file;

        if ($file != '' && $csv_separator != '' && $id_carrier !== 0) {
            return TRSCarrierTableRate::importCSV($file, $csv_separator, $id_carrier, $id_lang, $id_shop);
        }

        return array(
            'status' => 'danger',
            'message' => 'Error importing file'
        );
    }

    public function export()
    {
        $id_lang = $this->context->language->id;
        $id_shop = $this->context->shop->id;
        $id_carrier = (int)Tools::getValue('id_carrier', 0);
        $id_carrier = self::getCarrierIdFromCarrierId($id_carrier);
        $csv_separator = Tools::getValue('csv_separator', ',');

        if ($id_carrier !== 0) {
            TRSCarrierTableRate::exportCSV($id_carrier, $csv_separator, $id_lang, $id_shop);
        }

        die();
    }

    public function searchZones()
    {
        $search = Tools::getValue('q', '');
        $shop = $this->context->shop->id;

        $sql = new DbQuery();
        $sql->select('z.`id_zone` as `id`, z.`name` as `text`');
        $sql->from('zone', 'z');
        $sql->leftJoin('zone_shop', 'zs', 'zs.`id_zone` = z.`id_zone`');
        $sql->where('zs.`id_shop` = ' . $shop . ' AND `name` LIKE \'%' . $search . '%\' AND `active` = 1');
        $zones = Db::getInstance()->executeS($sql);

        return $zones;
    }

    public function searchCountries()
    {
        $search = Tools::getValue('q', '');
        $shop = $this->context->shop->id;
        $lang = $this->context->language->id;

        $sql = new DbQuery();
        $sql->select('c.`id_country` as `id`, cl.`name` as `text`');
        $sql->from('country', 'c');
        $sql->leftJoin('country_shop', 'cs', 'cs.`id_country` = c.`id_country`');
        $sql->leftJoin('country_lang', 'cl', 'cl.`id_country` = c.`id_country`');
        $sql->where('cs.`id_shop` = ' . $shop . ' AND cl.`id_lang` = ' . $lang . ' AND `name` LIKE \'%' . $search . '%\' AND `active` = 1');
        $countries = Db::getInstance()->executeS($sql);

        return $countries;
    }

    public function searchStates()
    {
        $search = Tools::getValue('q', '');
        $id_country = (int)Tools::getValue('c', 0);

        $sql = new DbQuery();
        $sql->select('`id_state` as `id`, `name` as `text`');
        $sql->from('state');
        $sql->where('`id_country` = ' . (int)$id_country . ' AND `name` LIKE \'%' . $search . '%\' AND `active` = 1');
        $states = Db::getInstance()->executeS($sql);

        return $states;
    }

    public function getProductsModel($id_lang = null, $id_category = null, $search = '', $start = 0, $limit = 10, $orderfld = 'p.`id_product`', $orderdir = 'ASC', $id_shop = null)
    {
        $search_filtered = AdminTableRateShippingController::filterSearchQuery($search, array('id', 'name'));

        if ($id_lang == null) {
            $id_lang = Context::getContext()->language->id;
        }

        $sql = new DbQuery();
        $sql->select('DISTINCT p.`id_product`');
        $sql->from('product', 'p');
        $sql->leftJoin('product_shop', 's', 'p.`id_product` = s.`id_product`');
        $sql->leftJoin('product_lang', 'l', 'p.`id_product` = l.`id_product`');

        if ($id_category != null) {
            $sql->leftJoin('category_product', 'cp', 'p.`id_product` = cp.`id_product`');
            $sql->where('cp.`id_category` = ' . (int)$id_category);
        }

        if ($search_filtered['id'] != '') {
            $sql->where('p.`id_product` IN (' . (int)$search_filtered['id'] . ')');
        }

        if ($search_filtered['name'] != '') {
            $sql->where('l.`name` LIKE \'%' . pSQL($search_filtered['name']) . '%\'');
        }

        if ($search_filtered['id'] == '' && $search_filtered['name'] == '' && $search_filtered['query'] != '') {
            $sql->where('l.`name` LIKE \'%' . pSQL($search_filtered['query']) . '%\'');
        }

        $sql->where('l.`id_lang` = ' . (int)$id_lang . ' AND s.`id_shop` = l.`id_shop`');

        if ($id_shop != null) {
            $sql->where('s.`id_shop` IN (' . (int)$id_shop . ')');
            $sql->where('l.`id_shop` IN (' . (int)$id_shop . ')');
        }

        $sql->limit($limit, $start);
        $sql->orderby($orderfld . ' ' . $orderdir);

        return Db::getInstance()->executeS($sql);
    }

    public function getProductsModelNumRows($id_category = null, $search = '', $id_shop = null)
    {
        $search_filtered = AdminTableRateShippingController::filterSearchQuery($search, array('id', 'name'));

        $sql = new DbQuery();
        $sql->select('COUNT(DISTINCT p.`id_product`) AS total');
        $sql->from('product', 'p');
        $sql->leftJoin('product_shop', 's', 'p.`id_product` = s.`id_product`');
        $sql->leftJoin('product_lang', 'l', 'p.`id_product` = l.`id_product`');

        if ($id_category != null) {
            $sql->leftJoin('category_product', 'cp', 'p.`id_product` = cp.`id_product`');
            $sql->where('cp.`id_category` = ' . (int)$id_category);
        }

        if ($search_filtered['id'] != '') {
            $sql->where('p.`id_product` IN (' . (int)$search_filtered['id'] . ')');
        }

        if ($search_filtered['name'] != '') {
            $sql->where('l.`name` LIKE \'%' . pSQL($search_filtered['name']) . '%\'');
        }

        if ($search_filtered['id'] == '' && $search_filtered['name'] == '' && $search_filtered['query'] != '') {
            $sql->where('l.`name` LIKE \'%' . pSQL($search_filtered['query']) . '%\'');
        }

        $sql->where('s.`id_shop` = l.`id_shop`');

        if ($id_shop != null) {
            $sql->where('s.`id_shop` IN (' . (int)$id_shop . ')');
            $sql->where('l.`id_shop` IN (' . (int)$id_shop . ')');
        }

        return (int)Db::getInstance()->getValue($sql);
    }

    public function hasProductCarrier($id_product, $id_carrier, $id_shop = null)
    {
        if ($id_shop === null) {
            $id_shop = $this->context->shop->id;
        }

        $sql = new DbQuery();
        $sql->select('COUNT(c.id_carrier) as count');
        $sql->from('product_carrier', 'pc');
        $sql->leftjoin('carrier', 'c', 'pc.id_carrier_reference = c.id_reference');
        $sql->where('deleted = 0 AND pc.id_product = ' . (int)$id_product . ' AND pc.id_shop = ' . (int)$id_shop . ' AND c.id_carrier = ' . (int)$id_carrier);

        return (int)((int)Db::getInstance()->getValue($sql) > 0);
    }

    public static function filterSearchQuery($search_query = '', $keys = array())
    {
        $search_query = trim($search_query);
        $search_items = array();

        foreach ($keys as $key) {
            $pattern = '/' . $key . '\s*:\s*[^:]*((?=' . implode('\s*:)|(?=', $keys) . '\s*:)|($))/i';
            preg_match($pattern, $search_query, $matches);

            if (isset($matches[0])) {
                $search_value = preg_replace('/\s+/', ' ', $matches[0]);
            } else {
                $search_value = '';
            }

            $search_value = explode(':', $search_value);
            $search_items[$key] = end($search_value);
            $search_items[$key] = trim($search_items[$key]);
        }

        $search_items['query'] = $search_query;

        return $search_items;
    }

    public static function getCarrierIdFromCarrierId($id_carrier)
    {
        $sql = 'SELECT `id_carrier`
                FROM `' . _DB_PREFIX_ . 'carrier`
                WHERE `id_reference` = (
                    SELECT `id_reference`
                    FROM `' . _DB_PREFIX_ . 'carrier`
                    WHERE `id_carrier` = ' . (int)$id_carrier . '
                ) AND `deleted` = 0';

        return (int)Db::getInstance()->getValue($sql);
    }
}