<?php
/**
 * Overrides carrier shipping with Table Rate Shipping
 *
 * Table Rate Shipping by Kahanit(https://www.kahanit.com/) is licensed under a
 * Creative Creative Commons Attribution-NoDerivatives 4.0 International License.
 * Based on a work at https://www.kahanit.com/.
 * Permissions beyond the scope of this license may be available at https://www.kahanit.com/.
 * To view a copy of this license, visit http://creativecommons.org/licenses/by-nd/4.0/.
 *
 * @author    Amit Sidhpura <amit@kahanit.com>
 * @copyright 2015 Kahanit
 * @license   http://creativecommons.org/licenses/by-nd/4.0/
 */

ini_set('max_execution_time', 0);

require_once(dirname(__FILE__) . '/autoload.php');
require_once(dirname(__FILE__) . '/StringExpressionLanguageProvider.php');
use Symfony\Component\ExpressionLanguage\ExpressionLanguage;

/**
 * Class TRSCarrierTableRate
 */
class TRSCarrierTableRate extends ObjectModel
{
    public $id_shop;

    public $id_carrier;

    public $order;

    public $active;

    public $id_zone;

    public $id_country;

    public $id_state;

    public $dest_city;

    public $dest_zip;

    public $condition_weight_from;

    public $condition_weight_to;

    public $condition_price_from;

    public $condition_price_to;

    public $condition_ptprice_from;

    public $condition_ptprice_to;

    public $condition_quantity_from;

    public $condition_quantity_to;

    public $condition_volume_from;

    public $condition_volume_to;

    public $price;

    public $comment;

    public static $error_message = array(
        'id_shop' => 'Invalid `Products`',
        'id_carrier' => 'Invalid `Carrier Id`',
        'id_zone' => 'Invalid `Zone`',
        'id_country' => 'Invalid `Country`',
        'id_state' => 'Invalid `State`',
        'dest_city' => 'Invalid `City`',
        'dest_zip' => 'Invalid `Zip/Post code`',
        'dest_zip_from' => 'Invalid `Zip/Post code from`',
        'dest_zip_to' => 'Invalid `Zip/Post code to`',
        'dest_zip_from_less_then_dest_zip_to' => '`Zip/Post code from` should be less than `Zip/Post code to`',
        'no_range_if_wildcard' => 'You cannot specify range if wildcard is used in zip/post code',
        'condition_weight_from_less_than_condition_weight_to' => '`Weight from` should be less than `Weight to`',
        'condition_price_from_less_than_condition_price_to' => '`Price from` should be less than `Price to`',
        'condition_ptprice_from_less_than_condition_ptprice_to' => '`Price(pretax) from` should be less than `Price(pretax) to`',
        'condition_quantity_from_less_than_condition_quantity_to' => '`Quantity from` should be less than `Quantity to`',
        'condition_volume_from_less_than_condition_volume_to' => '`Volume from` should be less than `Volume to`',
        'price' => 'Invalid `Price`',
        'comment' => 'Invalid `Comment`',
        'rule_exists' => 'Duplicate table rate found',
        'csv_invalid_format' => 'CSV not properly formatted',
        'no_records_found' => 'No records found'
    );

    public static $definition = array(
        'table' => 'carrier_table_rate',
        'primary' => 'id_carrier_table_rate',
        'multishop' => true,
        'fields' => array(
            'id_shop' => array('type' => self::TYPE_INT, 'validate' => 'isUnsignedInt'),
            'id_carrier' => array('type' => self::TYPE_INT, 'validate' => 'isUnsignedInt'),
            'id_zone' => array('type' => self::TYPE_INT, 'validate' => 'isUnsignedInt'),
            'id_country' => array('type' => self::TYPE_INT, 'validate' => 'isUnsignedInt'),
            'id_state' => array('type' => self::TYPE_INT, 'validate' => 'isUnsignedInt'),
            'dest_city' => array('type' => self::TYPE_STRING, 'validate' => 'isCityName', 'size' => 64),
            'dest_zip' => array('type' => self::TYPE_STRING, 'size' => 25),
            'condition_weight_from' => array('type' => self::TYPE_FLOAT),
            'condition_weight_to' => array('type' => self::TYPE_FLOAT),
            'condition_price_from' => array('type' => self::TYPE_FLOAT),
            'condition_price_to' => array('type' => self::TYPE_FLOAT),
            'condition_ptprice_from' => array('type' => self::TYPE_FLOAT),
            'condition_ptprice_to' => array('type' => self::TYPE_FLOAT),
            'condition_quantity_from' => array('type' => self::TYPE_FLOAT),
            'condition_quantity_to' => array('type' => self::TYPE_FLOAT),
            'condition_volume_from' => array('type' => self::TYPE_FLOAT),
            'condition_volume_to' => array('type' => self::TYPE_FLOAT),
            'price' => array('type' => self::TYPE_HTML),
            'comment' => array('type' => self::TYPE_STRING),
            'active' => array('type' => self::TYPE_BOOL, 'validate' => 'isBool'),
            'order' => array('type' => self::TYPE_INT, 'validate' => 'isUnsignedInt')
        )
    );

    public function validateFields($die = false, $error_return = true)
    {
        $message = parent::validateFields($die, $error_return);

        /* core validation */
        if ($message != 1) {
            foreach (TRSCarrierTableRate::$definition['fields'] as $field_name => $field_rules) {
                if (array_key_exists('validate', $field_rules)) {
                    preg_match('/->' . $field_name . ' /', $message, $matches);

                    if (count($matches) != 0) {
                        return TRSCarrierTableRate::$error_message[$field_name];
                    }
                }
            }
        }

        /* validate zone */
        if ($this->id_zone != 0 && $this->id_zone != '' && (!($zone = new Zone($this->id_zone)) || !Validate::isLoadedObject($zone))) {
            return TRSCarrierTableRate::$error_message['id_zone'];
        }

        /* validate country */
        if ($this->id_country != 0 && $this->id_country != '' && (!($country = new Country($this->id_country)) || !Validate::isLoadedObject($country))) {
            return TRSCarrierTableRate::$error_message['id_country'];
        }

        /* validate state */
        if ($this->id_state != 0 && $this->id_state != '' && ((!($state = new State($this->id_state)) || !Validate::isLoadedObject($state)) || $state->id_country != $this->id_country)) {
            return TRSCarrierTableRate::$error_message['id_state'];
        }

        /* validate zip codes */
        $dest_zip = explode('-', $this->dest_zip);
        $dest_zip_from = (isset($dest_zip[0])) ? trim($dest_zip[0]) : '';
        $dest_zip_from = ($dest_zip_from == '*') ? '' : $dest_zip_from;
        $dest_zip_to = (isset($dest_zip[1])) ? trim($dest_zip[1]) : '';
        $dest_zip_to = ($dest_zip_to == '*') ? '' : $dest_zip_to;
        $dest_zip_to = ($dest_zip_to == '') ? $dest_zip_from : $dest_zip_to;
        if (strcasecmp($dest_zip_from, $dest_zip_to) === 0) {
            $this->dest_zip = $dest_zip_from;
        } else {
            $this->dest_zip = $dest_zip_from . '-' . $dest_zip_to;
        }

        /* validate zip from */
        if ($dest_zip_from != '' && isset($country) && $country->zip_code_format && !self::checkZipCode($dest_zip_from,
                $country->zip_code_format, $country->iso_code)
        ) {
            return TRSCarrierTableRate::$error_message['dest_zip_from'];
        }

        /* validate zip to */
        if ($dest_zip_to != '' && isset($country) && $country->zip_code_format && !self::checkZipCode($dest_zip_to,
                $country->zip_code_format, $country->iso_code)
        ) {
            return TRSCarrierTableRate::$error_message['dest_zip_to'];
        }

        /* validate zip from and to range */
        if ($dest_zip_from != '' && $dest_zip_to != '' && $dest_zip_to < $dest_zip_from) {
            return TRSCarrierTableRate::$error_message['dest_zip_from_less_then_dest_zip_to'];
        }

        /* validate to cannot specify range if wildcard is used */
        if ($dest_zip_from != '' && $dest_zip_to != '' && (strpos($dest_zip_from, '*') != false || strpos($dest_zip_to,
                    '*') != false)
        ) {
            return TRSCarrierTableRate::$error_message['no_range_if_wildcard'];
        }

        /* validate condition value from and to range */
        $condition_types = ['weight', 'price', 'ptprice', 'quantity', 'volume'];
        foreach ($condition_types as $condition_type) {
            if ($this->{'condition_' . $condition_type . '_to'} < $this->{'condition_' . $condition_type . '_from'}) {
                return TRSCarrierTableRate::$error_message['condition_' . $condition_type . '_from_less_than_condition_' . $condition_type . '_to'];
            }
        }

        /* validate price */
        $vars = array(
            'tw' => 1,
            'tp' => 1,
            'tptp' => 1,
            'tq' => 1,
            'tv' => 1,
            'ctwf' => 1,
            'ctwt' => 1,
            'ctwi' => 1,
            'ctpf' => 1,
            'ctpt' => 1,
            'ctpi' => 1,
            'ctptpf' => 1,
            'ctptpt' => 1,
            'ctptpi' => 1,
            'ctqf' => 1,
            'ctqt' => 1,
            'ctqi' => 1,
            'ctvf' => 1,
            'ctvt' => 1,
            'ctvi' => 1,
            'skip' => -1
        );
        $this->price = str_replace('$', '', $this->price); // for bc with old formula variables with $
        $language = new ExpressionLanguage();
        $language->registerProvider(new StringExpressionLanguageProvider());
        try {
            $language->evaluate($this->price, $vars);
        } catch (Exception $e) {
            return TRSCarrierTableRate::$error_message['price'];
        }

        /* validate comment */
        $this->comment = str_replace('\"', '"', $this->comment);

        /* validate duplicate entry */
        $record = [
            'id_zone' => $this->id_zone,
            'id_country' => $this->id_country,
            'id_state' => $this->id_state,
            'dest_city' => $this->dest_city,
            'dest_zip' => $this->dest_zip,
            'condition_weight_from' => $this->condition_weight_from,
            'condition_weight_to' => $this->condition_weight_to,
            'condition_price_from' => $this->condition_price_from,
            'condition_price_to' => $this->condition_price_to,
            'condition_ptprice_from' => $this->condition_ptprice_from,
            'condition_ptprice_to' => $this->condition_ptprice_to,
            'condition_quantity_from' => $this->condition_quantity_from,
            'condition_quantity_to' => $this->condition_quantity_to,
            'condition_volume_from' => $this->condition_volume_from,
            'condition_volume_to' => $this->condition_volume_to,
            'price' => $this->price
        ];
        $rule_exists = self::ruleExists($this->id_carrier, $record, $this->id_shop);
        if ($rule_exists !== 0 && $rule_exists !== $this->id) {
            return TRSCarrierTableRate::$error_message['rule_exists'];
        }

        return $message;
    }

    /* rule group functions */

    public static function getRuleGroups($id_carrier, $id_zone = null, $start = null, $length = null, $where = '', $id_lang = null, $id_shop = null)
    {
        if ($id_lang === null) {
            $id_lang = Context::getContext()->language->id;
        }

        if ($id_shop === null) {
            $id_shop = Context::getContext()->shop->id;
        }

        $sql = new DbQuery();
        $sql->select('ct.id_zone, ct.id_country, ct.id_state, ct.dest_city,
            GROUP_CONCAT(DISTINCT ct.dest_zip ORDER BY ct.order ASC SEPARATOR \',\') as dest_zip,
            CONCAT(cl.name, \' (\', c.iso_code, \')\') as country, CONCAT(s.name, \' (\', s.iso_code, \')\') as state');
        $sql->from('carrier_table_rate', 'ct');
        $sql->leftJoin('country', 'c', 'c.id_country = ct.id_country');
        $sql->leftJoin('country_lang', 'cl', 'cl.id_country = ct.id_country AND cl.id_lang = ' . (int)$id_lang);
        $sql->leftJoin('country_shop', 'cs', 'cs.id_country = ct.id_country AND cs.id_shop = ' . (int)$id_shop);
        $sql->leftJoin('state', 's', 's.id_state = ct.id_state');
        $sql->where('ct.id_shop = ' . (int)$id_shop . ' AND ct.id_carrier = ' . (int)$id_carrier . (($id_zone !== null) ? ' AND ct.id_zone = ' . (int)$id_zone : ''));
        $sql->groupBy('ct.id_country, ct.id_state, ct.dest_city');
        $sql->orderBy('MIN(ct.order) ASC');

        if ($start !== null && $length !== null) {
            $sql->limit($length, $start);
        }

        $results = Db::getInstance()->executeS($sql);
        foreach ($results as &$result) {
            ($result['id_country'] == 0) ? $result['id_country'] = '*' : $result['id_country'];
            ($result['country'] == '') ? $result['country'] = '*' : $result['country'];
            ($result['id_state'] == 0) ? $result['id_state'] = '*' : $result['id_state'];
            ($result['state'] == '') ? $result['state'] = '*' : $result['state'];
            ($result['dest_city'] == '') ? $result['dest_city'] = '*' : $result['dest_city'];
            ($result['dest_zip'] == '') ? $result['dest_zip'] = '*' : $result['dest_zip'];
        }

        return $results;
    }

    public static function getRuleGroupsNumRows($id_carrier, $id_zone = null, $where = '', $id_shop = null)
    {
        if ($id_shop === null) {
            $id_shop = Context::getContext()->shop->id;
        }

        $sql = new DbQuery();
        $sql->select('ct.id_carrier_table_rate');
        $sql->from('carrier_table_rate', 'ct');
        $sql->where(
            (($id_shop) ? 'ct.id_shop = ' . (int)$id_shop . ' AND ' : '')
            . 'ct.id_carrier = ' . (int)$id_carrier
            . (($id_zone !== null) ? ' AND ct.id_zone = ' . (int)$id_zone : '')
        );
        $sql->groupBy('ct.id_country, ct.id_state, ct.dest_city');
        $sql_count = 'SELECT COUNT(id_carrier_table_rate) AS total FROM (' . $sql . ') AS tbl';

        return (int)Db::getInstance()->getValue($sql_count);
    }

    public static function ruleGroupExists($id_carrier, $id_zone, $id_country, $id_state, $dest_city, $id_shop = null)
    {
        $sql = new DbQuery();
        $sql->select('COUNT(id_carrier_table_rate) AS total');
        $sql->from('carrier_table_rate');
        $sql->where('id_carrier = ' . (int)$id_carrier . '
            AND id_zone = ' . (int)$id_zone . '
            AND id_country = ' . (int)(($id_country == '*') ? 0 : $id_country) . '
            AND id_state = ' . (int)(($id_state == '*') ? 0 : $id_state) . '
            AND dest_city = \'' . pSQL((($dest_city == '*') ? '' : $dest_city)) . '\'
            AND id_shop = ' . (int)$id_shop);

        return (int)Db::getInstance()->getValue($sql);
    }

    public static function getRuleGroupZips($id_carrier, $id_zone, $id_country, $id_state, $dest_city, $id_shop = null)
    {
        $sql = new DbQuery();
        $sql->select('DISTINCT dest_zip');
        $sql->from('carrier_table_rate');
        $sql->where('id_carrier = ' . (int)$id_carrier . '
            AND id_zone = ' . (int)$id_zone . '
            AND id_country = ' . (int)(($id_country == '*') ? 0 : $id_country) . '
            AND id_state = ' . (int)(($id_state == '*') ? 0 : $id_state) . '
            AND dest_city = \'' . pSQL((($dest_city == '*') ? '' : $dest_city)) . '\'
            AND id_shop = ' . (int)$id_shop);
        $sql->orderBy('`order` ASC');

        $results = Db::getInstance()->executeS($sql);
        $zips = [];
        if (is_array($results)) {
            foreach ($results as $result) {
                $zips[] = $result['dest_zip'];
            }
        }

        return $zips;
    }

    /* rule functions */

    public static function getRules($id_carrier, $id_zone, $id_country, $id_state, $dest_city, $start = null, $length = null, $where = '', $id_shop = null)
    {
        if ($id_shop === null) {
            $id_shop = Context::getContext()->shop->id;
        }

        $sql = new DbQuery();
        $sql->select('*, GROUP_CONCAT(`id_carrier_table_rate` ORDER BY `order` ASC SEPARATOR \',\') AS `id_carrier_table_rate`,
            GROUP_CONCAT(`dest_zip` ORDER BY `order` ASC SEPARATOR \',\') AS `dest_zip`');
        $sql->from('carrier_table_rate');
        $sql->where('
                `id_carrier` = ' . (int)$id_carrier . '
            AND `id_zone` = ' . (int)$id_zone . '
            AND `id_country` = ' . (int)(($id_country == '*') ? 0 : $id_country) . '
            AND `id_state` = ' . (int)(($id_state == '*') ? 0 : $id_state) . '
            AND `dest_city` = \'' . pSQL((($dest_city == '*') ? '' : $dest_city)) . '\'
            AND `id_shop` = ' . (int)$id_shop
        );
        $sql->groupBy('
            `condition_weight_from`, `condition_weight_to`,
            `condition_price_from`, `condition_price_to`,
            `condition_ptprice_from`, `condition_ptprice_to`,
            `condition_quantity_from`, `condition_quantity_to`,
            `condition_volume_from`, `condition_volume_to`,
            `price`, `active`'
        );
        $sql->orderBy('MIN(`order`) ASC');

        if ($start !== null && $length !== null) {
            $sql->limit($length, $start);
        }

        $results = Db::getInstance()->executeS($sql);
        $condition_types = ['weight', 'price', 'ptprice', 'quantity', 'volume'];
        foreach ($results as &$result) {
            ($result['id_country'] == 0) ? $result['id_country'] = '*' : $result['id_country'];
            ($result['id_state'] == 0) ? $result['id_state'] = '*' : $result['id_state'];
            ($result['dest_city'] == '') ? $result['dest_city'] = '*' : $result['dest_city'];
            ($result['dest_zip'] == '') ? $result['dest_zip'] = '*' : $result['dest_zip'];
            foreach ($condition_types as $condition_type) {
                ($result['condition_' . $condition_type . '_from'] == -1) ? $result['condition_' . $condition_type . '_from'] = '*' : $result['condition_' . $condition_type . '_from'];
                ($result['condition_' . $condition_type . '_to'] == -1) ? $result['condition_' . $condition_type . '_to'] = '*' : $result['condition_' . $condition_type . '_to'];
            }
        }

        return $results;
    }

    public static function getRulesNumRows($id_carrier, $id_zone, $id_country, $id_state, $dest_city, $where, $id_shop = null)
    {
        if ($id_shop === null) {
            $id_shop = Context::getContext()->shop->id;
        }

        $sql = new DbQuery();
        $sql->select('id_carrier_table_rate');
        $sql->from('carrier_table_rate');
        $sql->where('
                `id_carrier` = ' . (int)$id_carrier . '
            AND `id_zone` = ' . (int)$id_zone . '
            AND `id_country` = ' . (int)(($id_country == '*') ? 0 : $id_country) . '
            AND `id_state` = ' . (int)(($id_state == '*') ? 0 : $id_state) . '
            AND `dest_city` = \'' . pSQL((($dest_city == '*') ? '' : $dest_city)) . '\'
            AND `id_shop` = ' . (int)$id_shop
        );
        $sql->groupBy('
            `condition_weight_from`, `condition_weight_to`,
            `condition_price_from`, `condition_price_to`,
            `condition_ptprice_from`, `condition_ptprice_to`,
            `condition_quantity_from`, `condition_quantity_to`,
            `condition_volume_from`, `condition_volume_to`,
            `price`, `active`'
        );
        $sql_count = 'SELECT COUNT(id_carrier_table_rate) AS total FROM (' . $sql . ') AS tbl';

        return (int)Db::getInstance()->getValue($sql_count);
    }

    public static function ruleExists($id_carrier, $record, $id_shop = null)
    {
        $sql = new DbQuery();
        $sql->select('id_carrier_table_rate');
        $sql->from('carrier_table_rate');
        $sql->where('id_carrier = ' . (int)$id_carrier . '
            AND id_zone = ' . (int)$record['id_zone'] . '
            AND id_country = ' . (int)$record['id_country'] . '
            AND id_state = ' . (int)$record['id_state'] . '
            AND dest_city = \'' . pSQL($record['dest_city']) . '\'
            AND dest_zip = \'' . str_replace("'", "\'", $record['dest_zip']) . '\'
            AND condition_weight_from = ' . (float)$record['condition_weight_from'] . '
            AND condition_weight_to = ' . (float)$record['condition_weight_to'] . '
            AND condition_price_from = ' . (float)$record['condition_price_from'] . '
            AND condition_price_to = ' . (float)$record['condition_price_to'] . '
            AND condition_ptprice_from = ' . (float)$record['condition_ptprice_from'] . '
            AND condition_ptprice_to = ' . (float)$record['condition_ptprice_to'] . '
            AND condition_quantity_from = ' . (float)$record['condition_quantity_from'] . '
            AND condition_quantity_to = ' . (float)$record['condition_quantity_to'] . '
            AND condition_volume_from = ' . (float)$record['condition_volume_from'] . '
            AND condition_volume_to = ' . (float)$record['condition_volume_to'] . '
            AND price = ' . (float)$record['price'] . '
            AND id_shop = ' . (int)$id_shop);

        return (int)Db::getInstance()->getValue($sql);
    }

    public static function saveRules($action, $id_carrier, $records, $id_lang = null, $id_shop = null)
    {
        if ($id_shop === null) {
            $id_shop = Context::getContext()->shop->id;
        }
        if ($id_lang === null) {
            $id_lang = Context::getContext()->language->id;
        }
        $status = [
            'error' => 0,
            'success' => 0,
            'status' => '',
            'message' => []
        ];
        $fields = [
            'id_shop' => 0,
            'id_carrier' => 0,
            'id_zone' => 0,
            'id_country' => 0,
            'id_state' => 0,
            'dest_city' => '',
            'dest_zip' => '',
            'condition_weight_from' => -1,
            'condition_weight_to' => -1,
            'condition_price_from' => -1,
            'condition_price_to' => -1,
            'condition_ptprice_from' => -1,
            'condition_ptprice_to' => -1,
            'condition_quantity_from' => -1,
            'condition_quantity_to' => -1,
            'condition_volume_from' => -1,
            'condition_volume_to' => -1,
            'price' => '',
            'comment' => '',
            'active' => 0,
            'order' => 0
        ];
        $condition_types = ['weight', 'price', 'ptprice', 'quantity', 'volume'];
        $group_zips_sorted = [];
        $id_carrier_table_rate_saved = [];
        $groups_numrows = self::getRuleGroupsNumRows($id_carrier, null, '', $id_shop);

        if (is_array($records) && count($records) > 0) {
            foreach ($records as &$record) {
                // assign shop and carrier
                $record['id_shop'] = $id_shop;
                $record['id_carrier'] = $id_carrier;

                // prepare condition from and to values
                foreach ($condition_types as $condition_type) {
                    if (isset($record['condition_' . $condition_type])) {
                        if (is_numeric($record['condition_' . $condition_type])) {
                            $record['condition_' . $condition_type . '_from'] = (float)$record['condition_' . $condition_type];
                            $record['condition_' . $condition_type . '_to'] = (float)$record['condition_' . $condition_type];
                        } else {
                            $condition_values = explode('-', $record['condition_' . $condition_type]);
                            if (isset($condition_values[0]) && is_numeric($condition_values[0]) && isset($condition_values[1]) && is_numeric($condition_values[1])) {
                                $record['condition_' . $condition_type . '_from'] = (float)$condition_values[0];
                                $record['condition_' . $condition_type . '_to'] = (float)$condition_values[1];
                            }
                        }
                        unset($record['condition_' . $condition_type]);
                    }
                }

                // prepare record
                foreach ($fields as $field_key => $field_default) {
                    if (isset($record[$field_key])) {
                        $record[$field_key] = (is_numeric($field_default)) ? (float)$record[$field_key] : pSQL((($record[$field_key] == '*') ? '' : $record[$field_key]),
                            true);
                    } else {
                        $record[$field_key] = $field_default;
                    }
                }

                // prepare list of zipcodes
                if (isset($record['dest_zip'])) {
                    $dest_zips = (($record['dest_zip'] == '*') ? '' : $record['dest_zip']);
                } else {
                    $dest_zips = '';
                }
                $dest_zips = preg_replace('/[\\\r\\\n,]+/', '{{separator}}', $dest_zips);
                $dest_zips = explode('{{separator}}', $dest_zips);

                // get group rules
                $rules = TRSCarrierTableRate::getRules($record['id_carrier'], $record['id_zone'], $record['id_country'],
                    $record['id_state'], $record['dest_city'], null, null, '', $record['id_shop']);
                $group_zips = TRSCarrierTableRate::getRuleGroupZips($record['id_carrier'], $record['id_zone'],
                    $record['id_country'], $record['id_state'], $record['dest_city'], $record['id_shop']);

                // prepare delete rules with `group zips` not in `dest zips sent`
                $delete_zips = [];
                if ($action == 'edit') {
                    $delete_zips = array_diff($group_zips, $dest_zips);
                    if (count($delete_zips) > 0) {
                        $group_zips = array_diff($group_zips, $delete_zips);
                    }
                }

                // insert `all rules in database` for `dest zips sent`
                if (is_array($rules) && count($rules) > 0) {
                    foreach ($rules as $rule) {
                        foreach ($dest_zips as $dest_zip) {
                            // if dest zip is in database continue
                            if (in_array($dest_zip, $group_zips)) {
                                continue;
                            }

                            // setup object
                            $carrier_table_rate = new TRSCarrierTableRate();
                            foreach ($record as $record_key => $record_value) {
                                if (array_key_exists($record_key, $fields)) {
                                    $carrier_table_rate->{$record_key} = $record_value;
                                }
                            }

                            // assign rule data
                            foreach ($condition_types as $condition_type) {
                                $carrier_table_rate->{'condition_' . $condition_type . '_from'} = (($rule['condition_' . $condition_type . '_from'] == '*') ? -1 : $rule['condition_' . $condition_type . '_from']);
                                $carrier_table_rate->{'condition_' . $condition_type . '_to'} = (($rule['condition_' . $condition_type . '_to'] == '*') ? -1 : $rule['condition_' . $condition_type . '_to']);
                            }
                            $carrier_table_rate->price = $rule['price'];
                            $carrier_table_rate->comment = $rule['comment'];
                            $carrier_table_rate->active = $rule['active'];

                            // assign dest zip
                            $carrier_table_rate->dest_zip = $dest_zip;

                            // validate object
                            $message = $carrier_table_rate->validateFields(false, true);

                            // if duplicate record found continue
                            if (strcasecmp($message, self::$error_message['rule_exists']) === 0) {
                                continue;
                            }

                            // save object of validated successfully
                            if ($message == 1) {
                                $carrier_table_rate->save();
                                $carrier_table_rate->order = $carrier_table_rate->id;
                                $carrier_table_rate->save();

                                // save id for future delete if any error
                                $id_carrier_table_rate_saved[] = $carrier_table_rate->id;

                                $status['success']++;
                            } else {
                                $status['error']++;
                                $status['message'][] = $message;
                            }
                        }
                    }
                    $group_zips = array_unique(array_merge(array_diff($group_zips, $dest_zips), $dest_zips));
                } else {
                    $group_zips = $dest_zips;
                }

                // insert `rule sent` for `database dest zips`
                if (is_array($group_zips) && count($group_zips) > 0) {
                    foreach ($group_zips as $group_zip) {
                        // if group exists and sent rule is dummy continue
                        if (is_array($rules) && count($rules) > 0 && strcasecmp($record['comment'],
                                '{{dummy}}') === 0
                        ) {
                            continue;
                        }

                        // setup object
                        $carrier_table_rate = new TRSCarrierTableRate();
                        foreach ($record as $record_key => $record_value) {
                            if (array_key_exists($record_key, $fields)) {
                                $carrier_table_rate->{$record_key} = $record_value;
                            }
                        }

                        // assign dest zip
                        $carrier_table_rate->dest_zip = $group_zip;

                        // validate object
                        $message = $carrier_table_rate->validateFields(false, true);

                        // if duplicate record found continue
                        if (strcasecmp($message, self::$error_message['rule_exists']) === 0) {
                            continue;
                        }

                        // save object of validated successfully
                        if ($message == 1) {
                            $carrier_table_rate->save();
                            $carrier_table_rate->order = $carrier_table_rate->id;
                            $carrier_table_rate->save();

                            // save id for future delete if any error
                            $id_carrier_table_rate_saved[] = $carrier_table_rate->id;

                            $status['success']++;
                        } else {
                            $status['error']++;
                            $status['message'][] = $message;
                        }
                    }
                }

                // save group zips sort order for sorting
                $group_zips_sorted_key = $record['id_zone'] .
                    '{{separator}}' . (($record['id_country'] == 0) ? '*' : $record['id_country']) .
                    '{{separator}}' . (($record['id_state'] == 0) ? '*' : $record['id_state']) .
                    '{{separator}}' . (($record['dest_city'] == '') ? '*' : $record['dest_city']);
                $group_zips_sorted[$group_zips_sorted_key] = $group_zips;

                // execute delete rules with `group zips` not in `dest zips sent`
                if ($action == 'edit' && $status['error'] <= 0 && count($delete_zips) > 0) {
                    Db::getInstance()->delete('carrier_table_rate', 'id_carrier = ' . (int)$record['id_carrier'] . '
                    AND id_zone = ' . (int)$record['id_zone'] . '
                    AND id_country = ' . (int)$record['id_country'] . '
                    AND id_state = ' . (int)$record['id_state'] . '
                    AND dest_city = \'' . pSQL($record['dest_city']) . '\'
                    AND dest_zip IN (\'' . implode("','", $delete_zips) . '\')
                    AND id_shop = ' . (int)$id_shop);
                }

                // sort zones
                $sql = new DbQuery();
                $sql->select('DISTINCT id_zone');
                $sql->from('carrier_table_rate');
                $sql->where('id_carrier = ' . (int)$id_carrier . ' AND id_shop = ' . (int)$id_shop);
                $sql->orderBy('`order` ASC');
                $results = Db::getInstance()->executeS($sql);
                $zones = [];
                foreach ($results as $result) {
                    $zones[] = $result['id_zone'];
                }
                self::updateOrder('zone', $id_carrier, null, $zones);

                foreach ($zones as $id_zone) {
                    // sort rule groups
                    $results = TRSCarrierTableRate::getRuleGroups($id_carrier, $id_zone, null, null, '', $id_lang,
                        $id_shop);
                    $rulegroups = [];
                    foreach ($results as $result) {
                        $rulegroups[] = $result['id_country'] . '{{separator}}' . $result['id_state'] . '{{separator}}' . $result['dest_city'];
                    }
                    self::updateOrder('rulegroup', $id_carrier, $id_zone, $rulegroups);

                    // sort rules
                    foreach ($rulegroups as $rulegroup) {
                        $rulegroup_temp = explode('{{separator}}', $rulegroup);
                        $group_zips_sorted_key = $id_zone . '{{separator}}' . $rulegroup_temp[0] . '{{separator}}' . $rulegroup_temp[1] . '{{separator}}' . $rulegroup_temp[2];
                        if (!isset($group_zips_sorted[$group_zips_sorted_key])) {
                            $group_zips_sorted[$group_zips_sorted_key] = [];
                        }
                        $results = TRSCarrierTableRate::getRules($id_carrier, $id_zone, $rulegroup_temp[0],
                            $rulegroup_temp[1], $rulegroup_temp[2], null, null, '', $id_shop);
                        $rules = [];
                        foreach ($results as $result) {
                            $id_carrier_table_rate_explode = explode(',', $result['id_carrier_table_rate']);
                            $dest_zip_explode = explode(',', $result['dest_zip']);
                            $id_carrier_table_rate_explode_sorted = [];
                            foreach ($group_zips_sorted[$group_zips_sorted_key] as $group_zip) {
                                $id_carrier_table_rate_explode_sorted[] = $id_carrier_table_rate_explode[array_search($group_zip,
                                    $dest_zip_explode)];
                            }
                            $rules[] = implode(',', $id_carrier_table_rate_explode_sorted);
                        }
                        self::updateOrder('rule', null, null, $rules);
                    }
                }
            }
        }

        if ($status['error'] > 0) {
            $status['status'] = 'danger';
            $status['message'] = implode(', ', array_unique($status['message']));

            // delete saved records as error occured
            if (count($id_carrier_table_rate_saved) > 0) {
                Db::getInstance()->delete('carrier_table_rate',
                    'id_carrier_table_rate IN (' . implode(',', $id_carrier_table_rate_saved) . ')');
            }

            return $status;
        } else {
            $status['status'] = 'success';
            $status['message'] = 'Records saved successfully';

            // set carrier as table rate
            if ($groups_numrows == 0 && count($id_carrier_table_rate_saved) > 0) {
                self::setCarrierAsTableRate($id_carrier);
            }

            return $status;
        }
    }

    public static function updateOrder($entity = '', $id_carrier = null, $id_zone = null, $order = [], $id_shop = null)
    {
        if ($id_shop === null) {
            $id_shop = Context::getContext()->shop->id;
        }

        if ($entity !== '' && $order !== []) {
            switch ($entity) {
                case 'zone':
                    $order_ids = [];
                    foreach ($order as $order_item) {
                        $order_ids[] = (int)preg_replace('/^zone-/', '', $order_item);
                    }
                    $sql = 'SET @rank := 0;
                        UPDATE `' . _DB_PREFIX_ . 'carrier_table_rate`
                        SET `order` = @rank := @rank + 1
                        WHERE `id_carrier` = ' . (int)$id_carrier . ' AND `id_shop` = ' . (int)$id_shop . '
                        ORDER BY FIELD(`id_zone`, ' . implode(',', $order_ids) . '), `order` ASC';
                    Db::getInstance()->execute($sql);
                    break;
                case 'rulegroup':
                    if ($id_carrier !== 0 && $id_zone !== -1) {
                        $where = '`id_carrier` = ' . (int)$id_carrier . ' AND `id_zone` = ' . (int)$id_zone . ' AND `id_shop` = ' . (int)$id_shop . ' AND (';
                        foreach ($order as $order_item) {
                            $order_item_explode = explode('{{separator}}', $order_item);
                            $where .= '(`id_country` = ' . (int)(($order_item_explode[0] == '*') ? 0 : $order_item_explode[0]) . '
                                AND `id_state` = ' . (int)(($order_item_explode[1] == '*') ? 0 : $order_item_explode[1]) . '
                                AND `dest_city` = \'' . pSQL((($order_item_explode[2] == '*') ? '' : $order_item_explode[2])) . '\') OR ';
                        }
                        $where = preg_replace('/ OR $/', '', $where);
                        $where .= ')';

                        // get min order rank
                        $sql = new DbQuery();
                        $sql->select('MIN(`order`) AS `minorder`');
                        $sql->from('carrier_table_rate');
                        $sql->where($where);
                        $minorder = (int)Db::getInstance()->getValue($sql);
                        if (!$minorder) {
                            $minorder = 1;
                        }
                        $minorder--;

                        // update order
                        foreach ($order as $order_item) {
                            $order_item_explode = explode('{{separator}}', $order_item);

                            // prepare where clause
                            $where = '`id_carrier` = ' . (int)$id_carrier . '
                                AND `id_zone` = ' . (int)$id_zone . '
                                AND `id_country` = ' . (int)(($order_item_explode[0] == '*') ? 0 : $order_item_explode[0]) . '
                                AND `id_state` = ' . (int)(($order_item_explode[1] == '*') ? 0 : $order_item_explode[1]) . '
                                AND `dest_city` = \'' . pSQL((($order_item_explode[2] == '*') ? '' : $order_item_explode[2])) . '\'
                                AND `id_shop` = ' . (int)$id_shop;

                            // update order query
                            $sql = 'SET @rank := ' . (int)$minorder . ';
                                UPDATE `' . _DB_PREFIX_ . 'carrier_table_rate`
                                SET `order` = @rank := @rank + 1
                                WHERE ' . $where . '
                                ORDER BY `order` ASC';
                            Db::getInstance()->execute($sql);

                            // get no of rows updated
                            $sql = new DbQuery();
                            $sql->select('COUNT(`id_carrier_table_rate`) AS `total`');
                            $sql->from('carrier_table_rate');
                            $sql->where($where);
                            $row_count = (int)Db::getInstance()->getValue($sql);

                            $minorder += $row_count;
                        }
                    }
                    break;
                case 'rule':
                    $order_item_count = count(explode(',', reset($order)));
                    $order_temp = [];
                    for ($i = 0; $i < $order_item_count; $i++) {
                        foreach ($order as $order_item) {
                            $order_item_explode = array_map('intval', explode(',', $order_item));
                            $order_temp[] = $order_item_explode[$i];
                        }
                    }
                    $id_carrier_table_rate = implode(',', $order_temp);

                    // get min order rank
                    $sql = new DbQuery();
                    $sql->select('MIN(`order`) AS `minorder`');
                    $sql->from('carrier_table_rate');
                    $sql->where('`id_carrier_table_rate` IN (' . $id_carrier_table_rate . ')');
                    $minorder = (int)Db::getInstance()->getValue($sql);
                    if (!$minorder) {
                        $minorder = 1;
                    }
                    $minorder--;

                    // update order
                    $sql = 'SET @rank := ' . (int)$minorder . ';
                        UPDATE `' . _DB_PREFIX_ . 'carrier_table_rate`
                        SET `order` = @rank := @rank + 1
                        WHERE `id_carrier_table_rate` IN (' . $id_carrier_table_rate . ')
                        ORDER BY FIELD(`id_carrier_table_rate`, ' . $id_carrier_table_rate . '), `order` ASC';
                    Db::getInstance()->execute($sql);
                    break;
            }
        }

        return array(
            'status' => 'success',
            'message' => 'Order saved successfully'
        );
    }

    /* carrier functions */

    public static function setCarrierAsTableRate($id_carrier)
    {
        return Db::getInstance()->update('carrier', array(
            'is_module' => 1,
            'shipping_external' => 1,
            'need_range' => 1,
            'external_module_name' => 'tablerateshipping'
        ), 'id_carrier = ' . (int)$id_carrier);
    }

    public static function clearCarrierAsTableRate($id_carrier)
    {
        return Db::getInstance()->update('carrier', array(
            'is_module' => 0,
            'shipping_external' => 0,
            'need_range' => 0,
            'external_module_name' => ''
        ), 'id_carrier = ' . (int)$id_carrier);
    }

    public static function clearAllCarriersAsTableRate()
    {
        return Db::getInstance()->update('carrier', array(
            'is_module' => 0,
            'shipping_external' => 0,
            'need_range' => 0,
            'external_module_name' => ''
        ), 'external_module_name = \'tablerateshipping\'');
    }

    public static function updateIdCarrier($old_id_carrier = null, $new_id_carrier = null)
    {
        Db::getInstance()->update('carrier_table_rate', array(
            'id_carrier' => (int)$new_id_carrier
        ), 'id_carrier = "' . (int)$old_id_carrier . '"');
    }

    /* bulk functions */

    public static function bulkDelete($id_carrier_table_rate = array(), $delete_all = false, $id_carrier = null, $id_shop = null)
    {
        if ($id_shop === null) {
            $id_shop = Context::getContext()->shop->id;
        }

        if ($delete_all) {
            if ($id_carrier === null) {
                return false;
            } else {
                return Db::getInstance()->delete('carrier_table_rate',
                    'id_carrier = ' . (int)$id_carrier . ' AND id_shop = ' . (int)$id_shop);
            }
        } else {
            if (!is_array($id_carrier_table_rate) || count($id_carrier_table_rate) == 0) {
                return false;
            } else {
                return Db::getInstance()->delete('carrier_table_rate',
                    'id_carrier_table_rate IN (' . implode(',', array_map('intval', $id_carrier_table_rate)) . ')');
            }
        }
    }

    public static function importCSV($file, $csv_separator = ',', $id_carrier, $id_lang = null, $id_shop = null)
    {
        if ($id_lang === null) {
            $id_lang = Context::getContext()->language->id;
        }

        if ($id_shop === null) {
            $id_shop = Context::getContext()->shop->id;
        }

        $first_record = true;
        $valid_columns = [
            'zone',
            'country',
            'state',
            'city',
            'zip',
            'condition_weight',
            'condition_price',
            'condition_ptprice',
            'condition_quantity',
            'condition_volume',
            'price',
            'active',
            'comment'
        ];
        $column_numbers = [];
        $records = [];

        if (($handle = fopen($file, 'r')) !== false) {
            while (($data = fgetcsv($handle, 0, $csv_separator)) !== false) {
                if ($first_record) {
                    foreach ($data as $key => $value) {
                        if (in_array($value, $valid_columns)) {
                            $column_numbers[$value] = $key;
                        }
                    }

                    if (count($column_numbers) != count($valid_columns)) {
                        return array(
                            'status' => 'danger',
                            'message' => TRSCarrierTableRate::$error_message['csv_invalid_format']
                        );
                    }

                    $first_record = false;
                    continue;
                }

                $id_zone = (int)(($data[$column_numbers['zone']] == '*') ? 0 : Zone::getIdByName($data[$column_numbers['zone']]));
                $id_country = (int)(($data[$column_numbers['country']] == '*') ? 0 : Country::getByIso($data[$column_numbers['country']]));
                $id_state = (int)(($data[$column_numbers['state']] == '*') ? 0 : State::getIdByIso($data[$column_numbers['state']],
                    $id_country));

                $records[] = [
                    'id_zone' => $id_zone,
                    'id_country' => $id_country,
                    'id_state' => $id_state,
                    'dest_city' => pSQL(($data[$column_numbers['city']] == '*') ? '' : $data[$column_numbers['city']]),
                    'dest_zip' => pSQL(($data[$column_numbers['zip']] == '*') ? '' : preg_replace('/[\\\r\\\n,]+/',
                        '{{separator}}', $data[$column_numbers['zip']])),
                    'condition_weight' => pSQL(($data[$column_numbers['condition_weight']] == '*') ? -1 : $data[$column_numbers['condition_weight']]),
                    'condition_price' => pSQL(($data[$column_numbers['condition_price']] == '*') ? -1 : $data[$column_numbers['condition_price']]),
                    'condition_ptprice' => pSQL(($data[$column_numbers['condition_ptprice']] == '*') ? -1 : $data[$column_numbers['condition_ptprice']]),
                    'condition_quantity' => pSQL(($data[$column_numbers['condition_quantity']] == '*') ? -1 : $data[$column_numbers['condition_quantity']]),
                    'condition_volume' => pSQL(($data[$column_numbers['condition_volume']] == '*') ? -1 : $data[$column_numbers['condition_volume']]),
                    'price' => pSQL($data[$column_numbers['price']], true),
                    'active' => (int)$data[$column_numbers['active']],
                    'comment' => pSQL($data[$column_numbers['comment']])
                ];
            }

            fclose($handle);
        }

        return self::saveRules('add', $id_carrier, $records, $id_lang, $id_shop);
    }

    public static function exportCSV($id_carrier, $csv_separator = ',', $id_lang = null, $id_shop = null)
    {
        if ($id_lang === null) {
            $id_lang = Context::getContext()->language->id;
        }

        if ($id_shop === null) {
            $id_shop = Context::getContext()->shop->id;
        }

        $sql = new DbQuery();
        $sql->select('ct.*, GROUP_CONCAT(ct.dest_zip ORDER BY ct.order ASC SEPARATOR \'\n\') AS dest_zip,
            z.name as zone, cl.name as country, c.iso_code as c_iso_code, s.name as state, s.iso_code as s_iso_code');
        $sql->from('carrier_table_rate', 'ct');
        $sql->leftJoin('zone', 'z', 'ct.id_zone = z.id_zone');
        $sql->leftJoin('country', 'c', 'c.id_country = ct.id_country');
        $sql->leftJoin('country_lang', 'cl', 'cl.id_country = ct.id_country AND cl.id_lang = ' . (int)$id_lang);
        $sql->leftJoin('country_shop', 'cs', 'cs.id_country = ct.id_country AND cs.id_shop = ' . (int)$id_shop);
        $sql->leftJoin('state', 's', 's.id_state = ct.id_state');
        $sql->where('ct.id_carrier = ' . (int)$id_carrier);
        $sql->groupBy('
            ct.id_zone, ct.id_country, ct.id_state, ct.dest_city,
            ct.condition_weight_from, ct.condition_weight_to,
            ct.condition_price_from, ct.condition_price_to,
            ct.condition_ptprice_from, ct.condition_ptprice_to,
            ct.condition_quantity_from, ct.condition_quantity_to,
            ct.condition_volume_from, ct.condition_volume_to,
            ct.price, ct.active');
        $sql->orderBy('MIN(ct.order) ASC');

        $records = Db::getInstance()->executeS($sql);

        if (count($records) > 0) {
            $counter = 0;
            $collection = array();
            foreach ($records as $record) {
                $weight = ($record['condition_weight_from'] == $record['condition_weight_to']) ? $record['condition_weight_from'] : $record['condition_weight_from'] . '-' . $record['condition_weight_to'];
                $price = ($record['condition_price_from'] == $record['condition_price_to']) ? $record['condition_price_from'] : $record['condition_price_from'] . '-' . $record['condition_price_to'];
                $ptprice = ($record['condition_ptprice_from'] == $record['condition_ptprice_to']) ? $record['condition_ptprice_from'] : $record['condition_ptprice_from'] . '-' . $record['condition_ptprice_to'];
                $quantity = ($record['condition_quantity_from'] == $record['condition_quantity_to']) ? $record['condition_quantity_from'] : $record['condition_quantity_from'] . '-' . $record['condition_quantity_to'];
                $volume = ($record['condition_volume_from'] == $record['condition_volume_to']) ? $record['condition_volume_from'] : $record['condition_volume_from'] . '-' . $record['condition_volume_to'];

                $collection[$counter] = new stdClass();
                $collection[$counter]->zone = ($record['zone'] == '') ? '*' : $record['zone'];
                $collection[$counter]->country = ($record['c_iso_code'] == '') ? '*' : $record['c_iso_code'];
                $collection[$counter]->state = ($record['s_iso_code'] == '') ? '*' : $record['s_iso_code'];
                $collection[$counter]->city = ($record['dest_city'] == '') ? '*' : $record['dest_city'];
                $collection[$counter]->zip = ($record['dest_zip'] == '') ? '*' : $record['dest_zip'];
                $collection[$counter]->condition_weight = ($weight == -1) ? '*' : $weight;
                $collection[$counter]->condition_price = ($price == -1) ? '*' : $price;
                $collection[$counter]->condition_ptprice = ($ptprice == -1) ? '*' : $ptprice;
                $collection[$counter]->condition_quantity = ($quantity == -1) ? '*' : $quantity;
                $collection[$counter]->condition_volume = ($volume == -1) ? '*' : $volume;
                $collection[$counter]->price = $record['price'];
                $collection[$counter]->comment = $record['comment'];
                $collection[$counter]->active = $record['active'];
                $counter++;
            }
            $csv = new TRSCSV($collection, 'export', $csv_separator);
            $csv->export();
        } else {
            echo self::$error_message['no_records_found'];
        }

        die();
    }

    /* frontend functions */

    public static function getShippingPrice($id_carrier, $id_country, $id_state, $dest_city, $dest_zip, $vars, $id_shop = null)
    {
        $dest_zip = str_replace('-', '', str_replace(' ', '', $dest_zip));

        if ($id_shop === null) {
            $id_shop = Context::getContext()->shop->id;
        }

        $sql = new DbQuery();
        $sql->select('*, SUBSTRING_INDEX(`dest_zip`, "-", 1) AS `dest_zip_from`, SUBSTRING_INDEX(`dest_zip`, "-", -1) AS `dest_zip_to`');
        $sql->from('carrier_table_rate');
        $sql->where('
                `id_shop`     = ' . (int)$id_shop . '
            AND `id_carrier`  = ' . (int)$id_carrier . '
            AND (`id_country` = ' . (int)$id_country . ' OR `id_country` = 0)
            AND (`id_state`   = ' . (int)$id_state . '   OR `id_state`   = 0)
            AND (`dest_city`  = "' . pSQL($dest_city) . '"    OR `dest_city`  = "")
            AND ("' . (float)$vars['tw'] . '"   BETWEEN `condition_weight_from`   AND `condition_weight_to`   OR (`condition_weight_from`   < 0 AND `condition_weight_to`   < 0))
            AND ("' . (float)$vars['tp'] . '"   BETWEEN `condition_price_from`    AND `condition_price_to`    OR (`condition_price_from`    < 0 AND `condition_price_to`    < 0))
            AND ("' . (float)$vars['tptp'] . '" BETWEEN `condition_ptprice_from`  AND `condition_ptprice_to`  OR (`condition_ptprice_from`  < 0 AND `condition_ptprice_to`  < 0))
            AND ("' . (float)$vars['tq'] . '"   BETWEEN `condition_quantity_from` AND `condition_quantity_to` OR (`condition_quantity_from` < 0 AND `condition_quantity_to` < 0))
            AND ("' . (float)$vars['tv'] . '"   BETWEEN `condition_volume_from`   AND `condition_volume_to`   OR (`condition_volume_from`   < 0 AND `condition_volume_to`   < 0))
            AND `active` = 1
        ');
        $sql->having('
            (
                `dest_zip_from` = ""
                AND `dest_zip_to` = ""
            )
            OR (
                `dest_zip_from` = ""
                AND "' . pSQL($dest_zip) . '" LIKE REPLACE(REPLACE(REPLACE(`dest_zip_to`, "*", "%"), " ", ""), "-", "")
            )
            OR (
                "' . pSQL($dest_zip) . '" LIKE REPLACE(REPLACE(REPLACE(`dest_zip_from`, "*", "%"), " ", ""), "-", "")
                AND `dest_zip_to` = ""
            )
            OR (
                `dest_zip_from` != ""
                AND `dest_zip_to` != ""
                AND REPLACE(REPLACE(`dest_zip_from`, " ", ""), "-", "") <= "' . pSQL($dest_zip) . '"
                AND REPLACE(REPLACE(`dest_zip_to`, " ", ""), "-", "") >= "' . pSQL($dest_zip) . '"
            )
        ');
        $sql->orderBy('`order` ASC');
        $results = Db::getInstance()->executeS($sql);

        $language = new ExpressionLanguage();
        $language->registerProvider(new StringExpressionLanguageProvider());
        $shipping_price = false;

        foreach ($results as $result) {
            $vars_temp = $vars;
            $vars_temp = array_merge($vars_temp, [
                'ctwf' => $result['condition_weight_from'],
                'ctwt' => $result['condition_weight_to'],
                'ctwi' => $vars_temp['tw'] - $result['condition_weight_from'],
                'ctpf' => $result['condition_price_from'],
                'ctpt' => $result['condition_price_to'],
                'ctpi' => $vars_temp['tp'] - $result['condition_price_from'],
                'ctptpf' => $result['condition_ptprice_from'],
                'ctptpt' => $result['condition_ptprice_to'],
                'ctptpi' => $vars_temp['tptp'] - $result['condition_ptprice_from'],
                'ctqf' => $result['condition_quantity_from'],
                'ctqt' => $result['condition_quantity_to'],
                'ctqi' => $vars_temp['tq'] - $result['condition_quantity_from'],
                'ctvf' => $result['condition_volume_from'],
                'ctvt' => $result['condition_volume_to'],
                'ctvi' => $vars_temp['tv'] - $result['condition_volume_from']
            ]);

            $result['price'] = str_replace('$', '', $result['price']); // for bc with old formula variables with $

            try {
                $shipping_price = $language->evaluate($result['price'], $vars_temp);
            } catch (Exception $e) {
                $shipping_price = false;
            }

            if ($shipping_price < 0) {
                continue;
            } else {
                break;
            }
        }

        return $shipping_price;
    }

    /* helper functions */

    public static function checkZipCode($zip_code, $zip_code_format, $iso_code)
    {
        $iso_code_wildcard = preg_replace('/(.)/i', '([$1_%*]|)', $iso_code);

        $zip_regexp = '/^' . $zip_code_format . '$/ui';
        $zip_regexp = str_replace(' ', '( |)', $zip_regexp);
        $zip_regexp = str_replace('-', '(-|)', $zip_regexp);
        $zip_regexp = str_replace('N', '([0-9_%*]|)', $zip_regexp);
        $zip_regexp = str_replace('L', '([a-zA-Z_%*]|)', $zip_regexp);
        $zip_regexp = str_replace('C', $iso_code_wildcard, $zip_regexp);

        return (bool)preg_match($zip_regexp, $zip_code);
    }
}